# -*- coding: utf-8 -*-

import sys, pathlib, logging
from unittest import mock
import pytest #@UnusedImport #@UnresolvedImport
from library_main import check_and_configure_TEST # type: ignore
import library_main # type: ignore
import configure_logging # type: ignore
from concurrent_log_handler import ConcurrentRotatingFileHandler  # type: ignore
# from src.library_core.constants import *
from constants_lib import *

@pytest.mark.parametrize('production_run', [True, False])    
def test_sets_up_sys_production_run(production_run):
    if hasattr(sys, 'production_run'):
        delattr(sys, 'production_run')
    with mock.patch('library_core.configure_logging.configure_logging', side_effect=Exception):
        with pytest.raises(Exception):
            real_cwd = pathlib.Path.cwd()
            with mock.patch('pathlib.Path.cwd', return_value=pathlib.Path('blah', 'operative', 'blah') if production_run else real_cwd):
                check_and_configure_TEST()
    assert hasattr(sys, 'production_run') 
    assert sys.production_run == production_run

@pytest.mark.parametrize('log_config_success', [True, False])    
def test_logging_config_failure_ends_run(log_config_success, capsys):
    with mock.patch('library_core.configure_logging.configure_logging', return_value=log_config_success):
        with mock.patch('sys.exit') as mock_exit:
            with mock.patch('logging.getLogger', side_effect=Exception):
                with pytest.raises(Exception):
                    check_and_configure_TEST()
    assert mock_exit.call_count == (0 if log_config_success else 1)
    _, err = capsys.readouterr()
    assert ('FATAL' in err) != log_config_success 
    
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')                    
@pytest.mark.parametrize('get_admin_success', [True, False])    
def test_get_admin_success_failure_ends_run(get_admin_success, capsys):
    with mock.patch('sys.exit') as mock_exit:
        with mock.patch.object(library_main, 'get_is_admin', return_value=get_admin_success):
            sys.pytl.log(f'here get_admin_success {get_admin_success}')
            library_main.check_and_configure_TEST(admin_user=True)
    assert mock_exit.call_count == (0 if get_admin_success else 1)
    _, err = capsys.readouterr()
    assert ('FATAL' in err) != get_admin_success

@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')                
def test_initial_log_message_made(caplog):
    with mock.patch('library_core.configure_logging.configure_logging', return_value=True):
        with caplog.at_level(logging.INFO):
            library_main.check_and_configure_TEST()
    assert 'production run?' in caplog.records[0].message 
                
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')                
@pytest.mark.parametrize('os', ['win32', 'win64', 'linux', 'other_os'])
@mock.patch('library_core.configure_logging.configure_logging')
def test_os_is_windows_if_win_required(_, caplog, os):
    with mock.patch('sys.platform', new_callable=mock.PropertyMock(return_value=os)):
        with mock.patch('sys.argv', new_callable=mock.PropertyMock(return_value=('python', 'src/core'))):        
            if not os.startswith('win'):
                with caplog.at_level(logging.CRITICAL):
                    with pytest.raises(SystemExit):
                        library_main.check_and_configure_TEST(platform='win')
            else:
                library_main.check_and_configure_TEST(platform='win')
    if not os.startswith('win'):        
        assert 'FATAL' in caplog.text, f'caplog.text was |{caplog.text}|'
   
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')
@pytest.mark.parametrize('os', ['win32', 'win64', 'linux', 'other_os'])
# @pytest.mark.parametrize('os', ['win32'])
@mock.patch('library_core.configure_logging.configure_logging')
def test_by_default_win_or_lin_only_allowed(_, caplog, os):
    acceptable_os = os.startswith('win') or os.startswith('lin')
    with mock.patch('sys.platform', new_callable=mock.PropertyMock(return_value=os)):
        with mock.patch('sys.argv', new_callable=mock.PropertyMock(return_value=('python', 'src/core'))):        
            if not acceptable_os:
                with caplog.at_level(logging.CRITICAL):
                    with pytest.raises(SystemExit):
                        library_main.check_and_configure_TEST()
            else:
                library_main.check_and_configure_TEST()
    if acceptable_os:
        assert 'FATAL' not in caplog.text, f'caplog.text was |{caplog.text}|'
    else:        
        assert 'FATAL' in caplog.text, f'caplog.text was |{caplog.text}|'
   
@pytest.mark.parametrize('present', [True, False])
# @pytest.mark.parametrize('present', [True])
def test_c_and_c_checks_that_various_root_dir_files_exists(present, caplog, tmpdir):
    # filenames = ['start_ve.bat', '.gitignore', 'conftest.py', 'start_lin_ve']
    filenames = ['start_ve.bat', '.gitignore', 'conftest.py', 've_path_linux']
    tmpdir_path = pathlib.Path(str(tmpdir))
    if present:
        for filename in filenames:
            tmpdir_path.joinpath(filename).touch()
    with mock.patch('pathlib.Path.cwd', return_value=tmpdir_path):
        if not present:
            with caplog.at_level(logging.CRITICAL):
                with pytest.raises(SystemExit):
                    library_main.check_and_configure_TEST(platform='win')
            assert 'FATAL' in caplog.text, f'caplog.text was |{caplog.text}|'
        else:
            library_main.check_and_configure_TEST(platform='win')
            assert 'FATAL' not in caplog.text, f'caplog.text was |{caplog.text}|'
   
# # TODO: should be part of lib method checking for starter package... ("core" by default)  
# def test_core_package_exists(qtbot):
#     working_dir_path = pathlib.Path('src/core')
#     assert working_dir_path.is_dir()
#     file_path = pathlib.Path('src/core/__main__.py')
#     assert file_path.is_file()
# 
# # TODO: should be part of lib method for checking on installation of required modules
# def test_main_stops_if_packaging_module_not_installed(capsys, reset_packaging, qtbot):
#     # NB you have to have packaging installed in order to check the python version
#     # NB module "packaging" is often installed with other modules, so you may not have to do "pip install packaging"
#     if 'packaging' in sys.modules:
#         sys.modules['packaging'] = None
#     with pytest.raises(SystemExit):
#         with mock.patch('sys.argv', new_callable=mock.PropertyMock(return_value=('python', 'src/core'))):        
#             core_main.main()
#     _, err = capsys.readouterr()
#     assert 'FATAL' in err and 'pip install packaging' in err, f'err was |{err}|'
# 
# 
# # TODO: should be part of lib check method
# VersionInfo = collections.namedtuple('VersionInfo', 'major minor micro')     
# @pytest.mark.parametrize('version_found', [VersionInfo(3, 9, 1)])  
# @mock.patch('logging_config.configure_logger')
# @mock.patch('sys.argv', new_callable=mock.PropertyMock(return_value=('python', 'src/core')))        
# def test_main_stops_if_python_version_too_old(_, mock_config, capsys, version_found, qtbot):
#     sys.version = version_found
#     with mock.patch('sys.version_info', version_found):
#         if version_found == (2, 5, 2):
#             with pytest.raises(SystemExit):
#                 core_main.main()
#             _, err = capsys.readouterr()
#             assert 'FATAL' in err
#         else:
#             core_main.main() # other: passes OK!

def test_is_possible_to_configure_log_file_size(tmpdir):
    with mock.patch('concurrent_log_handler.ConcurrentRotatingFileHandler'):
        with pytest.raises(AssertionError):
            # passing "None" here will result in an AssertionError... but if the configure_logging() does not
            # have a default argument "maxBytes" it will raise a different one... so this guarantees that that param is present
            configure_logging.configure_logging(None, maxBytes = 1024 * 1024)
    

@pytest.mark.parametrize('maxBytes_size', [500, 100000, None])
def test_maxBytes_param_is_used_for_log_file_size(tmpdir, maxBytes_size):
    with mock.patch('concurrent_log_handler.ConcurrentRotatingFileHandler') as mock_constructor:
        with mock.patch('logging.Logger.info'):
            if maxBytes_size == None:
                configure_logging.configure_logging('some name')
            else:
                configure_logging.configure_logging('some name', maxBytes=maxBytes_size)
    # expected_path = pathlib.Path('D:/temp/logging/some name/rotator_fh.log')
    expected_path = pathlib.Path(f'{PART2_VAL}/temp/logging/some name/rotator_fh.log')
    if maxBytes_size == None:
        # this then checks on the default setting for maxBytes
        mock_constructor.assert_called_once_with(expected_path, 'a', 1024 * 1024, 10, 'utf-8')
    else:
        mock_constructor.assert_called_once_with(expected_path, 'a', maxBytes_size, 10, 'utf-8')




import logging
logging.basicConfig(format='%(asctime)s.%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
        datefmt='%H:%M:%S', level=logging.DEBUG)
basic_logger = logging.getLogger('BASIC')
basic_logger.info(f'start {__file__}')

import logging.handlers
import sys, logging, pathlib, subprocess, traceback
import colorlog # type: ignore 
from concurrent_log_handler import ConcurrentRotatingFileHandler  # type: ignore
import concurrent_log_handler # type: ignore

# 2026-01-07 YES! relative import 
from . constants_lib import *

STANDARD_LOGGING_LOCATION = pathlib.Path(f'{PART2_VAL}/{"ephem" if IS_PRODUCTION else "temp"}/logging')

def configure_logging(logger_name, caller_process='', use_colorlog=True, maxBytes = 1024 * 1024):
    assert type(logger_name) == str, f'|{logger_name}|: type is {type(logger_name)}'
    
    # you are not going to want to have multiple stream handlers... make a suitable mechanism
    # NB this mechanism means that the FIRST call to configure_logging will have the StreamHandler
    if not hasattr(logging, 'stream_handler'):
        if use_colorlog:
            stream_handler = colorlog.StreamHandler()
            # TODO the format pattern should include an indication about whether this is a production run or dev run...
            stream_handler.setFormatter(EnhancedColourFormatter(
                fmt='%(log_color)s%(asctime)s %(levelname)s %(module)s [%(pathname)s %(lineno)d]:\n%(message)s\n', 
                log_colors={
                    'DEBUG':    'white',
                    'INFO':     'white',
                    'WARNING':  'fg_bold_yellow',
                    'ERROR':    'fg_bold_red',
                    'CRITICAL': 'fg_bold_red,bg_white',
                }))
            logger = colorlog.getLogger(logger_name)
        else: 
            stream_handler = logging.StreamHandler(sys.stdout) # make it go to stdout
            stream_handler.setFormatter(EnhancedStdFormatter(
                fmt='%(asctime)s %(levelname)s %(module)s [%(pathname)s %(lineno)d]:\n%(message)s\n'))
            logger = logging.getLogger(logger_name)
        logger.addHandler(stream_handler)
        logging.stream_handler = True
    else:
        logger = colorlog.getLogger(logger_name) # has no StreamHandler
        """
    2023-06-18
    See my answer here: https://stackoverflow.com/questions/2266646/how-to-disable-logging-on-the-standard-error-stream/76499737#76499737
    To prevent "spurious" console output there are 3 possibilities: substitute logging.lastResort, set the parent of a logger
    to None, or give the root logger a handler which does not output to console.
        """
        
    # configure and add file-handler    
    generic_log_dir_path = pathlib.Path(STANDARD_LOGGING_LOCATION)
    """
2023-06-18
The logging framework *automatically* decides that a logger of name "xx.yy" is a child logger of a logger of name "xx".
Thus the 'parent' property of the former will be not the root logger, but logger "xx".

It is a good idea if the directory structure of the log files reflect this hierarchy.
if the name contains dots, this indicates a sequence of directories
e.g. "doc_indexer.user_logger" --> logging should be to directory ".../doc_indexer/user_logger/"
    """
    if '.' in logger_name:
        dirs = logger_name.split('.')
        specific_log_dir_path = generic_log_dir_path.joinpath(*dirs)
    else:
        specific_log_dir_path = generic_log_dir_path.joinpath(logger_name)
    log_file_path = specific_log_dir_path.joinpath(EnhancedStdFormatter.LOG_FILENAME)
    if not log_file_path.parent.is_dir():
        try:
            log_file_path.parent.mkdir(parents=True, exist_ok=True)
        except BaseException as e:
            msg = f"""{caller_process}WARNING: failed to create logging directory, |{log_file_path}|. 
No file-based logging will take place. Exception was |{e}| type {type(e)}"""
            raise Exception(msg)
            return
    file_handler = concurrent_log_handler.ConcurrentRotatingFileHandler(log_file_path, 'a', maxBytes, 10, 'utf-8')
    logger.addHandler(file_handler)
    file_formatter = EnhancedStdFormatter('%(asctime)s - [%(name)s] - %(levelname)s %(module)s [%(pathname)s %(lineno)d]:\n%(message)s')
    file_handler.setFormatter(file_formatter)
    logger.setLevel(logging.INFO)
    # message showing where the user can find the logger files (this message is also output to stdout)
    logger.info(f'{caller_process}Logger created, name |{logger.name}|. File handler log path: {log_file_path}')
    
    # 2026-01-07 needed! stops messages propagating up (and then down again? to another logger, e.g. basicConfig)
    logger.propagate = False

    return True

"""
2022-08-01
NB it is important to understand that these two logging handlers do not operate independently!
In fact they must be added in this order: 1) stream handler, then 2) file handler... and for example, in the
formatMessage method of the formatter of the first, it can be seen that record.pathname is set to a tweaked version 
of the path string... but this SAME record is then passed to the second handler, with this formatter: so the path 
has already been set to the "tweaked" version.
Also, the time format for the file handler's formatter must be "%(asctime)s", not "%(asctime)s.%(msecs)03d", as the 
format has already been set up for this record. Doing the above actually adds the ms string again. However, in this
second formatter you DO have to stipulate that a decimal point, not a comma, should be used for the ms part!     
"""
       
class EnhancedColourFormatter(colorlog.ColoredFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # print milliseconds after decimal point:
        self.default_msec_format = '%s.%03d'
    
    def formatMessage(self, record: logging.LogRecord):
        # tweak: curtail the pathname to show only the path elements from the CWD
        record_path = pathlib.Path(record.pathname)
        cwd_path = pathlib.Path.cwd()
        if record_path.is_relative_to(cwd_path):
            record.pathname = record_path.relative_to(cwd_path)
        # also search for logging being under "lib"
        # TODO do this properly TDD
        lib_dir_path = pathlib.Path(__file__).parent.parent.parent.parent
        if record_path.is_relative_to(lib_dir_path):
            record.pathname = record_path.relative_to(lib_dir_path)
        return super().formatMessage(record)
    
    def formatStack(self, stack_info):
        # 2025-01-22 added to remove the most egregious superfluous stack trace lines
        stack_info_lines = stack_info.splitlines()
        output_lines = []
        also_omit_next = False
        for stack_info_line in stack_info_lines:
            if also_omit_next:
                also_omit_next = False
                continue
            if stack_info_line.endswith('gui_checker_inner_function') or stack_info_line.endswith('executing_func'):
                also_omit_next = True
                continue
            output_lines.append(stack_info_line)
        return super().formatStack('\n'.join(output_lines).strip())

class EnhancedStdFormatter(colorlog.ColoredFormatter):
    LOG_FILENAME = 'rotator_fh.log'
    _LINE_OF_EQUALS = 120 * '='

    """
2022-08-01
NB it is important to understand that these two logging handlers do not operate independently!
In fact they must be added in this order: 1) stream handler, then 2) file handler... and for example, in the
formatMessage method of the formatter of the first, it can be seen that record.pathname is set to a tweaked version 
of the path string... but this SAME record is then passed to the second handler, with this formatter: so the path 
has already been set to the "tweaked" version.
Also, the time format for the file handler's formatter must be "%(asctime)s", not "%(asctime)s.%(msecs)03d", as the 
format has already been set up for this record. Doing the above actually adds the ms string again. However, in this
second formatter you DO have to stipulate that a decimal point, not a comma, should be used for the ms part!     
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # print milliseconds after decimal point:
        self.default_msec_format = '%s.%03d'

    def formatMessage(self, record: logging.LogRecord):
        # TODO code duplication, see above (added 2023-06-19)
        record_path = pathlib.Path(record.pathname)
        cwd_path = pathlib.Path.cwd()
        if record_path.is_relative_to(cwd_path):
            record.pathname = record_path.relative_to(cwd_path)
        # also search for logging being under "lib"
        # TODO do this properly TDD
        lib_dir_path = pathlib.Path(__file__).parent.parent.parent.parent
        if record_path.is_relative_to(lib_dir_path):
            record.pathname = record_path.relative_to(lib_dir_path)
        # super().formatMessage(record)
        # add lines above and below, and introduce newlines wherever "\\n" found in default output
        one_line_output = self._style.format(record)
        output = one_line_output.replace(r"\n", "\n") # substitute real newlines for newline escape sequences
        return f'{EnhancedStdFormatter._LINE_OF_EQUALS}\n{output}\n{EnhancedStdFormatter._LINE_OF_EQUALS}'
    
    def formatStack(self, stack_info):
        # 2025-01-22 added to remove the most egregious superfluous stack trace lines
        stack_info_lines = stack_info.splitlines()
        output_lines = []
        also_omit_next = False
        for stack_info_line in stack_info_lines:
            if also_omit_next:
                also_omit_next = False
                continue
            if stack_info_line.endswith('gui_checker_inner_function') or stack_info_line.endswith('executing_func'):
                also_omit_next = True
                continue
            output_lines.append(stack_info_line)
        return super().formatStack('\n'.join(output_lines).strip())




"""
2025-12-19
Purpose of this file: (NB TODO needs work...)
When setting up a new project, run this file in a given directory and, after making various checks, it will 
install the starting components for a new project. 

There are 4 prerequisites: 
- there must be no parameters, except "--h": help
- the python available must be >= 3.10 
- the OS must be Linux or W10
- the name (lower-cased) of the directory where this script is being run must include the string "workspace"

"""
import logging
logging.basicConfig(format='%(asctime)s.%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
        datefmt='%H:%M:%S', level=logging.DEBUG)
basic_logger = logging.getLogger('BASIC')
basic_logger.info(f'start {__file__}')

import sys
if len(sys.argv) > 1:
    helpscript = """
The purpose of this script is to set up basic utility files in a new (or existing) project directory and 
also to create a virtual environment where it should be (on "partition2"). It should work in Linux or W10,
and should create files which will mean the project can be run in either OS

**How to run**:
Run with no parameters, in a directory with "workspace" in the name (lower-cased): it will ask you the name 
of the project you want to create.
It will not interfere with expected files if it finds they already exist.
NB it assumes that "python" is a recognised executable in both Win and Lin: in Linux due to a symlink which 
may be local or system-wide (typically for /usr/bin/python3)
"""
    if len(sys.argv) == 2 and sys.argv[1] == '--h':
        basic_logger.info(helpscript)
    else:
        basic_logger.error(f'FATAL. This file must be run with no parameters or "--h"')
    sys.exit()

py_major_version = sys.version_info[0] 
py_minor_version = sys.version_info[1] 
if py_major_version < 3 or py_minor_version < 10:
    basic_logger.error(f'FATAL. This file must be run in Python 3.10 or higher')
    sys.exit()

platform_lc = sys.platform.lower()
if not platform_lc.startswith('lin') and not platform_lc.startswith('win'):
    basic_logger.error(f'FATAL. Cannot run this in OS |{sys.platform}| - only Linux or W10')
    sys.exit()

IS_LINUX = platform_lc.startswith('lin')
import pathlib
cwd = pathlib.Path.cwd()
if 'workspace' not in cwd.stem.lower():
    basic_logger.error(f'FATAL. This must be run in a directory containing the string "workspace"')
    sys.exit()

print(f'Enter project name:')
while True:
    project_name = input()
    if ' ' in project_name:
        basic_logger.warning(f'WARN. The project name cannot contain spaces. Try again')
    else:
        break

basic_logger.info(f'project name |{project_name}|')
# if this already exists as a file: error
target_root_dir_path = cwd.joinpath(project_name) 
if target_root_dir_path.is_file():
    basic_logger.error(f'FATAL. This |{project_name}| is a file in this directory')
    sys.exit()

# if this doesn't exist: make new directory
if target_root_dir_path.exists():
    basic_logger.info(f'Project {project_name}| already exists as directory... continuing to set up where expected files are missing')
    
else:

    basic_logger.info(f'... trying to create project root dir ... ')
    try:
        target_root_dir_path.mkdir()
    except BaseException as e:
        basic_logger.exception(f'FATAL. This |{project_name}| could not be created as as directory')
        sys.exit()
    basic_logger.info(f'project root dir created')

# path to venv must be PART2/apps/Python/virtual_envs/[directory]
import os
part2_env_var_name = 'PART2'
part2_env_var_val = os.environ.get(part2_env_var_name)
if part2_env_var_val is None:
    basic_logger.error(f'FATAL. No environment variable "{part2_env_var_name}" was found')
    sys.exit()

ve_part2_path_parts = ['apps', 'Python', 'virtual_envs']
ve_part2_path = pathlib.Path(f'{part2_env_var_val}/').joinpath(*ve_part2_path_parts)
if ve_part2_path.is_file():
    basic_logger.error(f'FATAL. Expected path to directory containing virtual envs is a file: {ve_part2_path}')
    sys.exit()
if not ve_part2_path.exists():
    # ... try to create
    try:
        ve_part2_path.mkdir(parents=True, exist_ok=True)
    except BaseException as e:
        basic_logger.exception(f'FATAL. Path to directory containing virtual envs could not be created')
        sys.exit()

# we need to use or create a new venv called dev_[xxx]
dev_ve_name = 'dev_' + project_name
target_dev_ve_path = ve_part2_path.joinpath(dev_ve_name)
if target_dev_ve_path.is_file():
    basic_logger.error(f'FATAL. Target path to directory to contain dev virtual env is a file: {target_dev_ve_path}')
    sys.exit()

if target_dev_ve_path.exists():
    basic_logger.info(f'A directory already exists at {target_dev_ve_path} - assuming this is a regular venv')
else:
    basic_logger.info(f'... trying to create the dev venv for this project at {target_dev_ve_path}\nCan take a few seconds...')
    from venv import create
    try:
        create(target_dev_ve_path, with_pip=True)
    except BaseException as e:
        basic_logger.exception(f'FATAL. Error creating venv at {target_dev_ve_path}')
        sys.exit()

# now create lin_ve_path
lin_path_str = f'${part2_env_var_name}/{"/".join(ve_part2_path_parts)}/{dev_ve_name}/bin/activate\n'
lin_ve_path_file_path = target_root_dir_path.joinpath('lin_ve_path')
if not lin_ve_path_file_path.exists():
    path_file_path_forwardslash_version = str(lin_ve_path_file_path).replace('\\', '/')
    with open(path_file_path_forwardslash_version, 'w') as f:
        target_ve_path_forwardslash_version = str(target_dev_ve_path).replace('\\', '/')
        f.write(lin_path_str)
    extra_msg = '' if IS_LINUX else ' - when in Linux make this executable' 
    basic_logger.info(f'Created start ve file for Lin at {lin_ve_path_file_path}{extra_msg}')
    if IS_LINUX:
        # ... make executable
        import subprocess
        subprocess.run(['chmod', 'u+x', lin_ve_path_file_path])

# now create win_ve_path.bat
win_path_str = f'%{part2_env_var_name}%/{"/".join(ve_part2_path_parts)}/{dev_ve_name}/Scripts/activate\n'\
    .replace('/', '\\')
win_ve_path_file_path = target_root_dir_path.joinpath('win_ve_path.bat')
if not win_ve_path_file_path.exists():
    path_file_path_backslash_version = str(win_ve_path_file_path).replace('/', '\\')
    with open(path_file_path_backslash_version, 'w') as f:
        target_ve_path_backslash_version = str(target_dev_ve_path).replace('/', '\\')
        f.write(win_path_str)
    basic_logger.info(f'Created start ve file for Win at {win_ve_path_file_path}')

# create the run_[xxx].py file
run_project_file_path = target_root_dir_path.joinpath(f'run_{project_name}.py')
if not run_project_file_path.exists():
    # NB this assumes that there is a "src" dir and under that a "core" dir and under that a file "start.py"
    # not always the case ... but can be tweaked, copied and tweaked, etc. as required
    run_text = """#!./python

import subprocess, pathlib, sys, os

try:
    platform_lc = sys.platform.lower()    
    if platform_lc.startswith('lin'):
        import tempfile
        with open('lin_ve_path', 'r') as f:
            contents = f.read()
        commands = [contents.strip(), f'python src/core/start.py']
        with tempfile.NamedTemporaryFile(suffix = '.sh', mode= 'w') as tmp:
            tmp.write('\n'.join(commands))
            tmp.flush()
            execute_file = ['/bin/bash', tmp.name]
            subprocess.run(['chmod', 'u+x', tmp.name], check=True)
            subprocess.run(execute_file, check=True)
    elif platform_lc.startswith('win'):
        subprocess.run(r'win_ve_path.bat && python src/core/start.py', universal_newlines=True, check=True)
    else:
        basic_logger.error(f'FATAL. Cannot be run in operating system {sys.platform}')
except BaseException as e:
    import logging
    basic_logger.error(f'FATAL. Exception: {e}')
    basic_logger.exception('')
"""
    with open(run_project_file_path, 'w') as f:
        f.write(run_text)
    if IS_LINUX:
        # ... make executable
        import subprocess
        subprocess.run(['chmod', 'u+x', lin_ve_path_file_path])
    else:
        basic_logger.info(f'TODO when in Linux this run file, {run_project_file_path}, should be made executable')

# make version.toml if it doesn't exist
version_toml_file_path = target_root_dir_path.joinpath(f'version.toml')
if not run_project_file_path.exists():
    with open(version_toml_file_path, 'w') as f:
        f.write('version="0.0.0"') # the first commit with a git tag will set this to something meaningful

# make pyt.bat (for Win) and pyt (for Lin)
pyt_bat_file_path = target_root_dir_path.joinpath(f'pyt.bat')
if not pyt_bat_file_path.exists():
    pyt_bat_text = """@echo off
REM this passes on the command line args (including "..\lib\pyt.py" to the python utility in project lib)
python ..\lib\pyt.py %*
"""
    with open(pyt_bat_file_path, 'w') as f:
        f.write(pyt_bat_text)

pyt_sh_file_path = target_root_dir_path.joinpath(f'pyt.sh')
if not pyt_sh_file_path.exists():
    pyt_sh_text = """# this passes on the command line args (including "..\lib\pyt.py" to the python utility in project lib)
python ../lib/pyt.py "$@"
"""
    with open(pyt_sh_file_path, 'w') as f:
        f.write(pyt_sh_text)
        
    if IS_LINUX:
        # ... make executable
        import subprocess
        subprocess.run(['chmod', 'u+x', pyt_sh_file_path])
    else:
        basic_logger.info(f'TODO when in Linux this pyt.sh file, {pyt_sh_file_path}, should be made executable')


# make config.toml, initially with one setting, the min "lib" (i.e. mrodent-dev-lib/mrodent-lib) version
config_toml_file_path = target_root_dir_path.joinpath(f'config.toml')
if not config_toml_file_path.exists():
    with open(config_toml_file_path, 'w') as f:
        f.write('min_lib_project_version="1.1.0"')

# finally, a starter .gitignore
gitignore_text = """
*.pyc 
*.txt 
requirements.txt 
**/tmp 
**/temp
/.settings
.project
.pydevproject
*.log
*.tmp
*.diff
/.pytest_cache/
*/target/*
**/target/**
"""
gitignore_file_path = target_root_dir_path.joinpath(f'.gitignore')
if not gitignore_file_path.exists():
    with open(gitignore_file_path, 'w') as f:
        f.write(gitignore_text)

"""
2025-12-19
NB currently this is agnostic about whether a .git dir or .git file (pointing to a remote repo) is present ...
"""



import logging
logging.basicConfig(format='%(asctime)s.%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
        datefmt='%H:%M:%S', level=logging.DEBUG)
# logging.basicConfig(format='line %(lineno)d ============> %(message)s',
#         datefmt='%H:%M:%S', level=logging.DEBUG)
basic_logger = logging.getLogger('BASIC')
basic_logger.info(f'start {__file__}')

import pathlib, sys, os

# get project name
cwd_path = pathlib.Path.cwd()
r"""
2025-12-28 typically found when lib (mrodent-dev-lib) is a pip-installed module:
__file__ |D:\apps\Python\virtual_envs\dev_doc_indexer\lib\site-packages\library_core\constants.py|
cwd_path |D:\My documents\software projects\EclipseWorkspace\doc_indexer|
"""
# get type of run
cwd_path_parts = cwd_path.parts
# determined_run_type = False
IS_PRODUCTION = None
run_type_index = -1
n_path_parts = len(cwd_path_parts)
for i, part in enumerate(cwd_path_parts):
  part_lc = part.lower()
  if part_lc == 'operative':
    # if being run from the project dir, "operative" must be the penultimate path part
    if i != n_path_parts - 2:
      basic_logger.error(f'FATAL. "operative" was found, but not as the penultimate CWD path part: {cwd_path_parts}')
      sys.exit()
    IS_PRODUCTION = True
    break
# NB this test is needed to stop the value of "i" increasing if IS_PRODUCTION is now True
if IS_PRODUCTION is None:
  for i, part in enumerate(cwd_path_parts):
    part_lc = part.lower()
    if 'workspace' in part_lc:
      # if being run from the project dir, "workspace" must be the penultimate path part
      if i != n_path_parts - 2:
        basic_logger.error(f'FATAL. a Path part containing "workspace" was found, but not as the penultimate CWD path part: {cwd_path_parts}')
        sys.exit()
      IS_PRODUCTION = False
      break
if IS_PRODUCTION is None:
  basic_logger.error(f'FATAL. {__file__}. Neither "operative" nor "workspace" found in CWD path parts: {cwd_path_parts}')
  sys.exit()

if not IS_PRODUCTION:
  basic_logger.info(f'--- this is a DEV run of file {__file__}')

basic_logger.info(f'+++ cwd_path_parts {cwd_path_parts} - i {i}')  

# the root directory (and thus project name will be the directory after the "operative" or "workspace" directory)
# and always as the final path part of CWD...
PROJECT_NAME = cwd_path_parts[-1]

# get PART2 env var value
PART2_NAME = 'PART2'
PART2_VAL = os.environ.get(PART2_NAME)
if PART2_VAL == None:
  basic_logger.error(f'FATAL. Environment variable |{PART2_NAME}| not set')
  sys.exit()

# get OS
lc_os_name = sys.platform.lower()
IS_LINUX = None
if lc_os_name.startswith('lin'):
  IS_LINUX = True
elif lc_os_name.startswith('win'):
  IS_LINUX = False
if IS_LINUX == None:
  basic_logger.error(f'FATAL. Cannot operate in OS {sys.platform}')
  sys.exit()



#!/usr/bin/env python3

import logging
logging.basicConfig(format='%(asctime)s.%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
        datefmt='%H:%M:%S', level=logging.DEBUG)
basic_logger = logging.getLogger('BASIC')
basic_logger.info(f'start {__file__}')

import pathlib, sys, os
import time, re, shutil, importlib, toml # type: ignore
import subprocess
from packaging import version as packaging_version

# from library_core.constants_lib import *
from . constants_lib import *

PRODUCTION_DIRECTORY_STR = f'{PART2_VAL}/My Documents/software projects/operative' 
REQUIRED_VERSION_REGEX_STR = r'\s*version\s*=\s*\'(\d+\.\d+\.\d+)\''

def main():
    # find where the call is coming from    
    cwd = pathlib.Path.cwd()
    # one of the ancestor directories of the cwd must contain the string "workspace"
    workspace_found = False
    for cwd_ancestor_dir in cwd.parts:
        cwd_ancestor_dir_lc = cwd_ancestor_dir.lower()
        if 'workspace' in cwd_ancestor_dir_lc:
            workspace_found = True
            break
        # but ... in fact if this script is being called from inside an installed package what you would hope to 
        # find is one of the parts being f"dev_{cwd.name}"
        if f'dev_{cwd.name}'.lower() in cwd_ancestor_dir_lc:
            workspace_found = True
            break
    if not workspace_found:
        basic_logger.error(f'FATAL. This script was not called from a development project directory: {cwd}')
        return    
    caller_proj_name = cwd.name
    basic_logger.info(f'Calling deploy for project |{cwd.name}|')
    dev_proj_dir_path = cwd
    # check that the calling project has a valid version file and version
    dev_version = get_version('development', dev_proj_dir_path)
    if dev_version == None: 
        return
    explanation = \
    """
    Explanation:
    What you should do therefore is create a new Virtual Environment with the name of this project but with the prefix "dev_...".
    Then you need to change the file which activates the VE (e.g. start_ve.bat).
    Then you will need to do "pip install ..." for all the required packages. That list can be obtained by starting the existing and 
    then installing them all (i.e. > pip freeze > requirement.txt).
    When the deployment takes place the activation file will be tweaked by the code to remove that "dev_" prefix, so don't delete the existing
    VE.
    Finally, if you have installed any new Python modules (ditto for Rust: Cargo.toml files, etc.) since the last deploy, the relevant upgrades 
    to the "run environment" will have to be made, i.e. for the "non-dev" VE, etc.
    """
    if IS_LINUX:
        # the file lin_ve_path should be tweaked: the content should be one line, 
        dev_lin_ve_file_path = dev_proj_dir_path.joinpath('lin_ve_path')
        with open(dev_lin_ve_file_path) as file:
            file_lines = list(file)
        if len(file_lines) != 1:
            basic_logger.error(f'FATAL. ve_path_linux should be one line but it has {len(file_lines)} lines')
            return
        expected_line = f'. ${PART2_NAME}/apps/Python/virtual_envs/dev_{caller_proj_name}/bin/activate' # should be using "development VE" in the dev location
        if file_lines[0].strip() != expected_line:
            explanation = \
            basic_logger.error(f'FATAL. ve_path_linux should contain this line:\n|{expected_line}|\n... but contains this line:\n|{file_lines[0]}|\n{explanation}')
            return
    else:
        # the file start_ve.bat should be tweaked: the content should be one line, 
        dev_start_ve_w10_file_path = dev_proj_dir_path.joinpath('start_ve.bat')
        with open(dev_start_ve_w10_file_path) as file:
            file_lines = list(file)
        if len(file_lines) != 1:
            basic_logger.error(f'FATAL. start_ve.bat should be one line but it has {len(file_lines)} lines')
            return
        expected_line = fr'%{PART2_NAME}%\apps\Python\virtual_envs\dev_{caller_proj_name}\Scripts\activate' # should be using "development VE" in the dev location
        if file_lines[0].strip() != expected_line:
            basic_logger.error(f'FATAL. start_ve.bat should contain this line:\n|{expected_line}|\n... but contains this line:\n|{file_lines[0]}|\n{explanation}')
            return
    prod_dir_path = pathlib.Path(PRODUCTION_DIRECTORY_STR)
    if not prod_dir_path.is_dir():
        basic_logger.error(f'FATAL. The string given for the production directory is not a directory: |{PRODUCTION_DIRECTORY_STR}|')
        return
    prod_proj_dir_path = prod_dir_path.joinpath(caller_proj_name)
    if prod_proj_dir_path.is_file():
        basic_logger.error(f'FATAL. The project production dir is a file: |{prod_proj_dir_path}|')
        return
    # if the proj production dir exists ...
    if prod_proj_dir_path.is_dir():      
        # ... try to get the version
        prod_version = get_version('production', prod_proj_dir_path)
        if prod_version == None: 
            basic_logger.error(f'FATAL. Could not get production version')
            return
        proceed_anyway = False
        if dev_version > prod_version:
            basic_logger.info('OK - dev version higher')
        else:
            msg = f'ANOMALY. The development version is {dev_version}. This must be higher than production version, but this is {prod_version}. Do you want to proceed anyway? [y/N]'
            reply = input(msg)
            proceed_anyway = reply.strip().lower() == 'y'
            if not proceed_anyway:
                return
            shutil.rmtree(prod_proj_dir_path)
        if not proceed_anyway:
            # try to rename the existing production directory to show the version: if failure, error and exit`
            renamed_prod_proj_dir_path = prod_proj_dir_path.parent.joinpath(f'{caller_proj_name}_{prod_version}')
            if renamed_prod_proj_dir_path.exists():
                basic_logger.error(f'FATAL. Can\'t rename production project directory as path to which it is to be renamed already exists as file or directory: {renamed_prod_proj_dir_path}')
                return
            try:
                prod_proj_dir_path.replace(renamed_prod_proj_dir_path)
            except PermissionError as e:
                basic_logger.error(f'FATAL. permission error: {e}. The production application may be running...')
                return
    else:
        basic_logger.info(f'prod_proj_dir_path {prod_proj_dir_path} is not a directory. prod_proj_dir_path.exists() {prod_proj_dir_path.exists()}')        
    # try to copy over the upgrade to the production environment
    shutil.copytree(dev_proj_dir_path, prod_proj_dir_path)
    # NB various things are not needed by the production version in the production location
    # delete "tests" dir if it exists
    tests_dir_path = prod_proj_dir_path.joinpath('tests')
    if tests_dir_path.is_dir():
        shutil.rmtree(tests_dir_path)
    # delete ".git" dir if it exists
    git_dir_path = prod_proj_dir_path.joinpath('.git')
    if git_dir_path.is_dir():
        shutil.rmtree(git_dir_path)
    # delete .settings, etc.
    settings_dir_path = prod_proj_dir_path.joinpath('.settings')
    if settings_dir_path.is_dir():
        shutil.rmtree(settings_dir_path)
    for path in prod_proj_dir_path.glob('pyt.*'):
        path.unlink()
    for path in prod_proj_dir_path.glob('*.txt'):
        path.unlink()
    for path in prod_proj_dir_path.glob('.*project'):
        path.unlink()
    for path in prod_proj_dir_path.glob('.git*'):
        path.unlink()

    if IS_LINUX:
        basic_logger.info(f'caller_proj_name |{caller_proj_name}|')
        prod_lin_ve_file_path = prod_proj_dir_path.joinpath('lin_ve_path')
        with open(prod_lin_ve_file_path, 'w') as file:
            file.write(f'. {PART2_VAL}/apps/Python/virtual_envs/{caller_proj_name}/bin/activate') # will use the "production VE" for production runs
    else:
        prod_start_ve_w10_file_path = prod_proj_dir_path.joinpath('start_ve.bat')
        with open(prod_start_ve_w10_file_path, 'w') as file:
            file.write(f'{PART2_VAL}/apps/Python/virtual_envs/{caller_proj_name}/Scripts/activate') # will use the "production VE" for production runs

    """
    now that the production and dev projects use different VEs, these VEs will in fact be using different
    compiled Rust modules. "> maturin develop --release" (or "cargo build" for non-PyO3 modules) must therefore be run in 
    the production location.
    But in fact I've also decided that the production and dev projects should also have separate Rust "targets"
    ... this means that in dev populate_index_module/.cargo/config.toml must contain the following:
    [build]
    target-dir = ".../apps/rust/auto_generated/dev_doc_indexer_populate_index_module/target"
    ... and that in production populate_index_module/.cargo/config.toml must contain the following:
    [build]
    target-dir = ".../apps/rust/auto_generated/doc_indexer_populate_index_module/target"

    ... to make this as universal as possible, it is essential to "walk" through all directories from the root:
    - if the name of the dir is .cargo
    - and if there is a file config.toml there
    - and if there is a section "build"
    - and if there is a line "target_dir"
    - then that line should match pattern ".../apps/rust/auto_generated/dev_[root_dir].*/target" - if not, WARN
    - otherwise, in the copied over file, change to ".../apps/rust/auto_generated/[root_dir].*/target"
    - also run "> maturin develop --release"/"cargo build" in that directory

    """
    warnings_reported = False
    for path in prod_proj_dir_path.rglob("*"):
        if not path.is_dir(): continue
        if path.name != '.cargo': continue
        config_toml_file_path = path.joinpath('config.toml')
        if not config_toml_file_path.is_file(): 
            basic_logger.warning(f'\nWARN. {config_toml_file_path} is not a file')
            warnings_reported = True            
            continue
        try:
            configs = toml.load(config_toml_file_path)
        except BaseException as e:
            basic_logger.warning(f'\nWARN. toml.load failure on {config_toml_file_path}: {e}')
            warnings_reported = True            
            continue
        if 'build' not in configs:
            basic_logger.warning(f'\nWARN. {config_toml_file_path} does not have key "build"')
            warnings_reported = True            
            continue
        if 'target-dir' not in configs['build']:
            basic_logger.warning(f'\nWARN. {config_toml_file_path}["build"] does not have key "target-dir"')
            warnings_reported = True            
            continue
        pattern = f'{PART2_VAL}/apps/rust/auto_generated/dev_{caller_proj_name}_(.*)/target'
        target_dir = configs['build']['target-dir']
        match = re.fullmatch(pattern, target_dir)
        if match == None:
            basic_logger.warning(f'\nWARN. {config_toml_file_path} target-dir |{target_dir}| does not match pattern')
            warnings_reported = True            
            continue
        production_target = f'{PART2_VAL}/apps/rust/auto_generated/{caller_proj_name}_{match.groups()[0]}/target'
        configs['build']['target-dir'] = production_target
        try:
            with open(config_toml_file_path, 'w') as file:
                toml.dump(configs, file)
        except BaseException as e:
            warnings_reported = True            
            basic_logger.warning(f'\nWARN. toml.dump failure on {config_toml_file_path}: {e}')
            continue
        # you now have to run either "maturin develop --release" (if pyo3 project) or "cargo build"
        # so you have to find the Cargo.toml file and find whether one of the dependencies is "pyo3"
        cargo_toml_file_path = path.parent.joinpath('Cargo.toml')
        if not cargo_toml_file_path.is_file(): 
            warnings_reported = True            
            basic_logger.warning(f'\nWARN. {cargo_toml_file_path} is not a file')
        try:
            cargo_configs = toml.load(cargo_toml_file_path)
        except BaseException as e:
            warnings_reported = True            
            basic_logger.warning(f'\nWARN. toml.load failure on {cargo_toml_file_path}: {e}')
            continue
        if 'dependencies' in cargo_configs and 'pyo3' in cargo_configs['dependencies']:
            basic_logger.info(f'running "maturin develop --release" in dir {path.parent}')
            # now run the maturin command in dir path.parent
            completed_process = subprocess.run('maturin develop --release', cwd=path.parent)
        else:
            basic_logger.info(f'running "cargo build" in dir {path.parent}')
            completed_process = subprocess.run('cargo build', cwd=path.parent)
        if completed_process.returncode == 0:
            basic_logger.info('... process seems to have completed ok')
        else:
            warnings_reported = True            
            basic_logger.warning(f'WARN. Non-zero return code from process: {completed_process.returncode}')
    # "unlink" (i.e. delete) certain files in the production location
    run_deploy_file_path = prod_proj_dir_path.joinpath('run_deploy.py')
    # if run_deploy_file_path.is_file():
    run_deploy_file_path.unlink()

    msg = 'Standard deploy concluded'
    if warnings_reported:
        msg += ' but some warnings reported: see warnings'
    basic_logger.info(f'{"="*50}\n{msg}\n: project {caller_proj_name} version {dev_version} is now in production\n{"="*50}')
    # if there is a custom_deploy.py file in the calling root directory, run that file
    proj_cust_deploy_file_path = dev_proj_dir_path.joinpath('custom_deploy.py')
    if proj_cust_deploy_file_path.is_file():
        basic_logger.info(f'about to run custom deploy file: {proj_cust_deploy_file_path}')
        """
NB 
1. for importlib.import_module to work the module string given must point to a module when added 
to one of the entries in sys.path. The same applies when using "import xxx".
2. if you add a pathlib.Path to sys.path this won't work! It must be a string, i.e. of the directory where
your module is to be found.
3. when you go "import xxx", this does indeed run (execute) that file!
4. it's probably a good idea to remove this added entry in sys.path once you've finished importing. 
        """
        sys.path.append(str(cwd))
        import custom_deploy # type: ignore
        sys.path.remove(str(cwd))
        custom_deploy.main()
    # suggest a change to the version number in the development location
    msg = 'DEPLOY finished OK'
    if warnings_reported:
        msg += ' but some warnings reported: see warnings'
        
def get_version(location, proj_dir_path, old_style=False):
    # 2024-04-20 there are now two possibilities for the version file: 
    # 1) [project root dir]/version.toml (PREFERRED, TAKE PRECEDENCE) and 2) [project root dir]/src/version.py (old style)
    if not old_style:
        proj_version_file_path = proj_dir_path.joinpath('version.toml')
        if not proj_version_file_path.is_file():
            basic_logger.warning(f'WARN. {proj_version_file_path} is not a file. Will try to see if version can be obtained from version_file.py... ')
            return get_version(location, proj_dir_path, True)
        try:
            with open(proj_version_file_path, 'r') as f:
                toml_data = toml.load(f) # is a dict
            min_lib_version = toml_data['version']
            version = packaging_version.Version(min_lib_version)
            return version
        except BaseException as e:
            return get_version(location, proj_dir_path, True)
    proj_version_file_path = proj_dir_path.joinpath('src', 'version_file.py')
    if not proj_version_file_path.is_file():
        basic_logger.error(f'FATAL. Couldn\'t find the project {location} version file at: |{proj_version_file_path}|')
        return
    # take all the lines in the file and disregard all which are empty or starting with "\s*#"...
    version_file_lines = proj_version_file_path.read_text().splitlines()
    version_file_lines[:] = [line for line in version_file_lines if line != '' and not re.match(r'\s*#.*', line)]    
    if len(version_file_lines) == 0:
        basic_logger.error(f'FATAL. Project {location} version file at: |{proj_version_file_path}| contains no non-comment lines')
        return
    if len(version_file_lines) > 1:
        basic_logger.error(f'FATAL. Project {location} version file at: |{proj_version_file_path}| contains too many non-comment lines. ' + 
              'Should contain one line, like this: "version=\'0.0.2\'"')
        return
    # if there is a production app already in use, check the version
    match = re.match(REQUIRED_VERSION_REGEX_STR, version_file_lines[0])
    if not match:
        basic_logger.error(f'FATAL. Project {location} version file at: |{proj_version_file_path}| line ' + 
              f'Does not match required regex |{REQUIRED_VERSION_REGEX_STR}|')
        return
    # NB match.groups() will be a one-tuple
    try:
        version_to_return = packaging_version.parse(match.groups()[0])
        basic_logger.info(f'lib - version_to_return {version_to_return}')
        return version_to_return
    except packaging_version.InvalidVersion:
        basic_logger.error(f'FATAL. Project {location} version file seems to contain invalid version. Check file contents. ' + 
            f'Should contain one non-comment line, matching regex |{REQUIRED_VERSION_REGEX_STR}|')
        return
    
def revert(dev_proj_dir_path, prod_proj_dir_path):
    # TODO 2025-12-28: do I want to implement a proper "revert" function???
    pass
#     # NB it is possible that the original deploy operation was the first one done, so no previous version will exist in the production location
#     print('revert...')
#     # delete the current production project dir
#     if not prod_proj_dir_path.is_dir():
#         print(f'FATAL. During revert: the production project path {prod_proj_dir_path} is not a directory')
#         return
#     # find the version of this file you are just about to delete
#     prod_version = get_version('production', prod_proj_dir_path)
#     if prod_version == None:
#         return
#     # find all directories matching "{project_name}_\d+\.\d+.\d+" in the production directory
#     caller_proj_name = dev_proj_dir_path.name
#     production_dir_path = prod_proj_dir_path.parent
#     # map_retired_version_to_path = {}
#     newest_retired_version = None
#     newest_retired_version_path = None
#     for f in production_dir_path.iterdir():
#         if f.is_dir():
#             match = re.match(fr'{caller_proj_name}_(\d+\.\d+\.\d+)', f.name)
#             if match:
#                 dir_version = packaging_version.parse(match.groups()[0])
#                 # map_retired_version_to_path[dir_version] = f
#                 if newest_retired_version == None or newest_retired_version < dir_version:
                    
#                     # the one thing that a genuine "retired version directory" must contain is a "src" subfolder
#                     if not f.joinpath('src').is_dir():
#                         print(f'WARNING. The directory {f} should be a retired project production directory. ' + 
#                             'But it is excluded from consideration as it does not contain a "src" subdirectory')
#                     else:
#                         newest_retired_version = dir_version
#                         newest_retired_version_path = f
                        
#     # at least one such "retired version" found...
#     if newest_retired_version == None:
#         print(f'FATAL. No retired production version of project {caller_proj_name} was found, so cannot revert.')
#         return
#     print(f'newest_retired_version {newest_retired_version} path {newest_retired_version_path}')
#     # check that the version file in this "highest version dir path"
#     retired_version = get_version('newest retired production project', newest_retired_version_path)
#     if retired_version == None: return
#     # check that the production version is higher than the retired version. If not, error and exit
#     if prod_version <= retired_version:
#         print(f'FATAL. The production version is {prod_version} and the retired version is {retired_version}. ' + 
#             'But to do a revert the production version must be higher than the retired version.')
#         return 
#     # only at this point, having ascertained that there is at least one qualifying "retired" directory 
#     # for the project in the production directory, can we actually delete the current, operational, production directory
#     try:
#         shutil.rmtree(prod_proj_dir_path)
#     except PermissionError as e:
#         print(f'FATAL. permission error during revert: {e}. The production application may be running...')
#         return
#     if prod_proj_dir_path.exists():
#         print(f'FATAL. Although the production directory at {prod_proj_dir_path} should have been deleted, for some reason it still exists.')
#         return
#     # we now need to rename the latest retired directory to simple "{project_name}" 
#     try:
#         newest_retired_version_path.replace(prod_proj_dir_path)
#     except PermissionError as e:
#         print(f'FATAL. permission error: {e}')
#         return
#     print(f'Standard revert concluded: project {caller_proj_name} version {retired_version} has now been returned to production')
#     # if there is a custom_deploy.py file in the calling root directory, run that file
#     proj_cust_deploy_file_path = dev_proj_dir_path.joinpath('custom_deploy.py')
#     if proj_cust_deploy_file_path.is_file():
#         print(f'about to run custom deploy file with REVERT=True: {proj_cust_deploy_file_path}')
#         sys.path.append(str(dev_proj_dir_path))
#         import custom_deploy
#         sys.path.remove(str(dev_proj_dir_path))
#         custom_deploy.main(revert=True)
#     # suggest a change to the version number in the development location
#     print(f'REVERT finished OK. You should now change the version number in the dev project\'s version file')


# 2025-12-28 the name, if called by running "import library_core.deploy", will not be "__main__"
main()





import logging
logging.basicConfig(format='%(asctime)s.%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
        datefmt='%H:%M:%S', level=logging.DEBUG)
basic_logger = logging.getLogger('BASIC')
basic_logger.info(f'start {__file__}')

import logging, pathlib, sys, collections, re, importlib, copy
from unittest import mock
import toml # type: ignore
import packaging.version, pytest, colorlog 
from concurrent_log_handler import ConcurrentRotatingFileHandler  
# import src.library_core.library_main as library_main
import configure_logging #@UnusedImport #@UnresolvedImport
import library_main
from common_for_tests import AnyStringWithRegex, is_match

cwd_path = pathlib.Path.cwd()

@pytest.fixture(autouse=True)
def mock_reset_test_logger_situation():
    test_logger = logging.getLogger('test')
    assert len(test_logger.handlers) == 0
    test_existing_handlers = []
    for handler in test_logger.handlers:
        test_existing_handlers.append(handler)
    yield
    # restore pytest logger handlers (essential for any tests which test logging configuration!)     
    test_logger.handlers = test_existing_handlers

@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')
def test_version_file_contains_version():
    with open('version.toml', 'r') as f:
        version_dict = toml.load(f)
    assert 'version' in version_dict
    version_val_as_str = version_dict['version']
    from packaging import version as packaging_version #@UnusedImport #@UnresolvedImport
    version = packaging_version.Version(version_val_as_str)
    assert type(version) == packaging_version.Version, f'type is {type(version)}'

@pytest.mark.parametrize('os', ['win32', 'win64', 'linux', 'other_os'])
@pytest.mark.parametrize('return_vals', [True, False])
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
def test_os_is_windows_or_linux(_, capsys, os, return_vals):
    assert len(logging.getLogger().handlers) == 4    
    with mock.patch('sys.platform', new_callable=mock.PropertyMock(return_value=os)):
        with mock.patch('sys.argv', new_callable=mock.PropertyMock(return_value=('python', 'src/core'))):        
            if os == 'other_os' and not return_vals:
                with pytest.raises(SystemExit) as excinfo:
                    app_name = cwd_path.parts[-1]
                    return_val = library_main.do_start_run_checks(app_name, cwd_path.joinpath('src', 'core', '__main__.py'), return_value=return_vals)
                assert excinfo.value.code == library_main.UNKNOWN_OS                    
            else:
                app_name = cwd_path.parts[-1]
                return_val = library_main.do_start_run_checks(app_name, cwd_path.joinpath('src', 'core', '__main__.py'), return_value=return_vals)        
    _, err = capsys.readouterr()
    if os == 'other_os':
        if return_vals:
            assert return_val[0] == library_main.UNKNOWN_OS and re.match(r'FATAL\..*windows or linux platforms.*', return_val[1]) != None
        else:
            assert 'FATAL' in err, f'err was |{err}|'
    elif return_vals:
        assert return_val == library_main.CHECKS_PASSED  
 
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')
def test_start_ve_files_exists():
    # start_ve.bat is the file that activates the virtual env in W10, start_lin_ve in Linux
    file_path_str = 'start_ve.bat' if sys.platform.startswith('win') else 'start_lin_ve'
    assert pathlib.Path(file_path_str).is_file()
 
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above') 
def test_git_ignore_file_exists():
    file_path = pathlib.Path('.gitignore')
    assert file_path.is_file()
 
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above') 
def test_core_package_exists():
    working_dir_path = pathlib.Path('src/library_core')
    assert working_dir_path.is_dir()
    file_path = pathlib.Path('src', 'library_core', 'library_main.py')
    assert file_path.is_file()
 
@pytest.mark.parametrize('version_found', ['real', (3, 9, 1)])
@pytest.mark.parametrize('return_vals', [True, False])  
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
def test_min_python_version_v10(_, capsys, version_found, return_vals):
    if version_found == 'real':
        version_info = sys.version_info
    else:
        VersionInfo = collections.namedtuple('VersionInfo', 'major minor micro')
        version_info = VersionInfo(*version_found)
    with mock.patch('sys.argv', new_callable=mock.PropertyMock(return_value=('python', 'src/core'))):
        with mock.patch('sys.version_info', version_info):
            if (version_found == 'real') or return_vals:
                app_name = cwd_path.parts[-1]
                return_val = library_main.do_start_run_checks(app_name, cwd_path.joinpath('src', 'core', '__main__.py'), return_value=return_vals)        
            else:
                with pytest.raises(SystemExit) as excinfo:
                    app_name = cwd_path.parts[-1]
                    return_val = library_main.do_start_run_checks(app_name, cwd_path.joinpath('src', 'core', '__main__.py'), return_value=return_vals)
                assert excinfo.value.code == library_main.PYTHON_VERSION_IS_TOO_LOW                            
    _, err = capsys.readouterr()
    if return_vals:
        if version_found == 'real':
            assert return_val == library_main.CHECKS_PASSED
        else:
            assert return_val[0] == library_main.PYTHON_VERSION_IS_TOO_LOW and re.match( r'FATAL\..*Python 10 or higher.*', return_val[1]) != None 
    else:
        if version_found == 'real':
            assert 'FATAL' not in err, f'err was |{err}|'
            assert return_val == library_main.CHECKS_PASSED
        else:
            assert 'FATAL' in err, f'err was |{err}|'
 
@pytest.mark.parametrize('good_path', [True, False])
@mock.patch(f'pathlib.Path.cwd')
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
@pytest.mark.parametrize('return_vals', [True, False])  
def test_main_stops_if_not_run_from_top_directory(_, mock_cwd, good_path, capsys, return_vals):
    # NB this test (like others) only works if "pytest" is run from the project directory
    # NB assumption: that there is, above this file, a dir "tests", of which the parent dir is the project root dir
    file_path = pathlib.Path(__file__)
    while str(file_path.parts[-1]).lower() != 'tests':
        file_path = file_path.parent
    root_dir_path_str = file_path.parent.parts[-1].lower()
    if cwd_path.parts[-1] != root_dir_path_str:
        raise Exception(f'This test file must be run from the root project directory "{root_dir_path_str}", but cwd is {cwd_path}')
    if good_path:
        # a "good path" must end with the directory named [app_name]
        mock_cwd.return_value = cwd_path
        app_name = cwd_path.parts[-1]
        return_value = library_main.do_start_run_checks(app_name, mock_cwd.return_value.joinpath('src', 'core', '__main__.py'), return_value=return_vals)
        assert return_value ==  library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    else:
        mock_cwd.return_value = pathlib.Path('random_path')
        if return_vals:
            app_name = cwd_path.parts[-1]
            return_value = library_main.do_start_run_checks(app_name, mock_cwd.return_value, return_value=return_vals)        
            _, err = capsys.readouterr()
            assert return_value[0] == library_main.NOT_RUN_FROM_PROJECT_ROOT
            assert re.match( r'FATAL\..*top-level directory.*', return_value[1]) != None
            assert err == ''
        else:
            with pytest.raises(SystemExit) as excinfo:
                app_name = cwd_path.parts[-1]
                library_main.do_start_run_checks(app_name, mock_cwd.return_value, return_value=return_vals)        
            _, err = capsys.readouterr()
            assert re.match( r'FATAL\..*top-level directory.*', err) != None
            assert excinfo.value.code == library_main.NOT_RUN_FROM_PROJECT_ROOT
         
@pytest.mark.parametrize('good_path', [True, False])
@mock.patch(f'pathlib.Path.resolve')    
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
@pytest.mark.parametrize('return_vals', [True, False])      
def test_main_path_is_relative_to_cwd(_, mock_resolve, good_path, capsys, return_vals):
    file_path = pathlib.Path(__file__)
    while str(file_path.parts[-1]).lower() != 'tests':
        file_path = file_path.parent
    root_dir_path = file_path.parent
    if good_path:
        mock_resolve.return_value = root_dir_path.joinpath('src', 'core', '__main__.py')
        app_name = cwd_path.parts[-1]
        return_value = library_main.do_start_run_checks(app_name, mock_resolve.return_value)
        assert return_value ==  library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    else:
        app_name = file_path.parent.parts[-1].lower()
        # this "resolved" Path ends correctly, but the path to __main__.py is not under the cwd...
        caller_file_path = pathlib.Path('some_dir', app_name, 'src', 'core', '__main__.py')
        # to use these paths in a regex the backslashes need to be doubled in w10
        caller_file_path_str = str(caller_file_path).replace("\\", "\\\\")
        cwd_path_str = str(cwd_path).replace("\\", "\\\\")
        pattern_str = fr'FATAL\..*caller file {caller_file_path_str} is not under cwd ({cwd_path_str}).*'
        mock_resolve.return_value = caller_file_path
        if return_vals:
            app_name = cwd_path.parts[-1]
            return_val = library_main.do_start_run_checks(app_name, mock_resolve.return_value, return_value=return_vals)        
            _, err = capsys.readouterr()
            assert re.match(pattern_str, return_val[1], flags=re.IGNORECASE) != None
            assert return_val[0] == library_main.CALLER_FILE_NOT_UNDER_CWD
        else:
            with pytest.raises(SystemExit) as excinfo:
                app_name = cwd_path.parts[-1]
                library_main.do_start_run_checks(app_name, mock_resolve.return_value, return_value=return_vals)        
            _, err = capsys.readouterr()
            assert re.match(pattern_str, err, flags=re.IGNORECASE) != None
            assert excinfo.value.code == library_main.CALLER_FILE_NOT_UNDER_CWD

@pytest.mark.skip(reason="""because as of 2023-03-12 this test no longer applies: i.e. it is permitted to do these checks from a .py file
which is not in the package src/core of the calling package""")
@pytest.mark.parametrize('good_path', [True, False])
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
@pytest.mark.parametrize('return_vals', [True, False]) 
def test_main_path_relative_to_cwd_is_in_expected_package(_, good_path, capsys, return_vals):
    # NB this test started in a project other than lib... it is now slightly puzzling because 
    # another project will be calling do_start_run_checks()... but the main thing is to ensure that 
    # (unless an optional keyword "package_name" param is supplied) the path ends "src/core/[...some file]"
    file_path = pathlib.Path(__file__)
    while str(file_path.parts[-1]).lower() != 'tests':
        file_path = file_path.parent
    root_dir_path = file_path.parent
    if good_path:
        app_name = cwd_path.parts[-1]
        return_value = library_main.do_start_run_checks(app_name, root_dir_path.joinpath('src', 'core', '__main__.py'), return_value=return_vals)
        assert return_value ==  library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    else:
        caller_file_path = root_dir_path.joinpath('other_source_folder', 'core', '__main__.py')
        # to use these paths in a regex the backslashes need to be doubled in w10
        caller_file_path_str = str(caller_file_path).replace("\\", "\\\\")
        pattern_str = f'FATAL. This parent dirs of file {caller_file_path_str} must be src/core'
        if return_vals:
            app_name = cwd_path.parts[-1]
            return_val = library_main.do_start_run_checks(app_name, caller_file_path, return_value=return_vals)
            _, err = capsys.readouterr()
            assert re.match(pattern_str, return_val[1], flags=re.IGNORECASE) != None
            assert return_val[0] == library_main.CALLER_FILE_NOT_IN_EXPECTED_PACKAGE
        else:
            with pytest.raises(SystemExit) as excinfo:
                app_name = cwd_path.parts[-1]
                library_main.do_start_run_checks(app_name, caller_file_path, return_value=return_vals)
            _, err = capsys.readouterr()
            assert re.match(pattern_str, err, flags=re.IGNORECASE) != None
            assert excinfo.value.code == library_main.CALLER_FILE_NOT_IN_EXPECTED_PACKAGE
 
# NB 2021-12-22 this does actually need to configure the logger, so do not use fixture mock_out_configure_logger_method
# 2022-07-31 far from sure what the purpose of this test is!
def test_get_logger_should_deliver_colorlog_logger_with_at_least_one_handler(capsys):
    assert len(logging.getLogger().handlers) == 4
    app_name = cwd_path.parts[-1]
    configure_logging.configure_logging(app_name)
    root_logger = colorlog.getLogger()
    assert len(root_logger.handlers) > 0
    local_logger = colorlog.getLogger(__name__)
    assert len(local_logger.handlers) == 0
    main_logger = colorlog.getLogger('__main__')
    assert len(main_logger.handlers) == 0
    random_logger = colorlog.getLogger('???')
    assert len(random_logger.handlers) == 0
 
def test_logger_prints_to_stdout(capsys, tmpdir):
    configure_logging.configure_logging('test')
    logger = logging.getLogger('test')
    # sys.pytl.log(f'logger {logger}')
    
    # NB by default the logger is set at "WARNING or higher": this is changed by logger.setLevel(logging.DEBUG)
    logger.info('messageXXX')
    # NB TODO "warning" or higher should be printed to stderr (and be printed in bright red, even in W10); below that to stdout
    logger.warning('messageYYY')
    _, err = capsys.readouterr()
    assert 'messageXXX' in err and 'messageYYY' in err, f'err was |{err}|'
 
def test_configure_logging_creates_logger_with_info_level():
    assert len(logging.getLogger('test').handlers) == 0
    configure_logging.logger = None
    configure_logging.configure_logging('test')
    logger = logging.getLogger('test')
    assert logger.level == logging.INFO
    # this should create 2 handlers: stream handler (for stdout) and file handler (for file logging)
    assert len(logging.getLogger('test').handlers) == 2
 
def test_configure_logging_creates_and_adds_streamhandler_with_EnhancedColourFormatter():
    assert len(logging.getLogger().handlers) == 4
    configure_logging.configure_logging('test')
    # NB 2 (I think) of pytest's only loggers are colorlog.StreamHandlers!
    n_stream_handlers_with_ecf = 0
    logger = logging.getLogger('test')
#     sys.pytl.log(f'logger {logger}')
#     sys.pytl.log(f'len(logger.handlers) {len(logger.handlers)}')
    for handler in logger.handlers:
        if isinstance(handler, colorlog.StreamHandler):
            if isinstance(handler.formatter, configure_logging.EnhancedColourFormatter):
                logging.info(f'handler.formatter {handler.formatter}')
                n_stream_handlers_with_ecf += 1
                formatter = handler.formatter
    assert n_stream_handlers_with_ecf == 1
    # sys.pytl.log(f'formatter._fmt |{formatter._fmt}|')
    assert is_match(r'^%\(log_color\)s%\(asctime\)s %\(levelname\)s %\(module\)s \[%\(pathname\)s %\(lineno\)d\]:\n%\(message\)s\n$', formatter._fmt)
 
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above') 
def test_EnhancedColourFormatter_prints_only_path_from_cwd(capsys):
    sys.pytl.log(f'hasattr(logging, "stream_handler") {hasattr(logging, "stream_handler")}')
    configure_logging.configure_logging('test')
    logger = logging.getLogger('test')
    sys.pytl.log(f'hasattr(logging, "stream_handler") {hasattr(logging, "stream_handler")}')
    logger.info('some bubbles')
    _, err = capsys.readouterr()
    # the date-time part at the start (after the day of the week) should not contain a comma
    # this output string starts with several colour-control characters and then the day of the week
    module_name = __name__
    pattern_str = fr'.*INFO {module_name} \[\s*(.+{module_name}.py) \d+\]:.*some bubbles.*'
    # sys.pytl.log(f'out |{out}|')
    result = re.match(pattern_str, err, re.DOTALL)
    assert result != None
    file_path = pathlib.Path(__file__)
    last_part_of_path = file_path.relative_to(cwd_path.parent)
    assert result.groups()[0] == str(last_part_of_path)
 
def test_configure_logging_creates_and_adds_ConcurrentRotatingFileHandler_with_standard_formatter():
    assert len(logging.getLogger().handlers) == 4
    assert configure_logging.configure_logging('test')
    n_crf_handlers_with_std_formatter = 0
    
    # for handler in configure_logging.logger.handlers:
    logger = logging.getLogger('test')    
    for handler in logger.handlers:
        
        if isinstance(handler, ConcurrentRotatingFileHandler):
            basic_logger.info(f'dir(handler) {dir(handler)}')
            if isinstance(handler.formatter, logging.Formatter):
                n_crf_handlers_with_std_formatter += 1
    assert n_crf_handlers_with_std_formatter == 1
 
def test_ConcurrentRotatingFileHandler_is_created_with_certain_params():
    assert len(logging.getLogger().handlers) == 4
    with mock.patch('concurrent_log_handler.ConcurrentRotatingFileHandler.__init__', return_value=None) as mock_constructor:
        with mock.patch('logging.Logger.info'):
            configure_logging.configure_logging('test')
    logging_file_path = pathlib.Path(configure_logging.STANDARD_LOGGING_LOCATION, 'test', configure_logging.EnhancedStdFormatter.LOG_FILENAME)    
    # mock_constructor.assert_called_once_with(logging_file_path, 'a', 100000, 10, 'utf-8')
    # change to default size 2024-12-07
    mock_constructor.assert_called_once_with(logging_file_path, 'a', 1024 * 1024, 10, 'utf-8')
 
def test_file_formatter_is_created_with_certain_params():
    assert len(logging.getLogger().handlers) == 4
    
    # with mock.patch('logging.Formatter.__init__', return_value=None) as mock_constructor:
    with mock.patch.object(configure_logging, 'EnhancedStdFormatter', return_value=None) as mock_constructor:
        
        with mock.patch('logging.Logger._log'):
            configure_logging.configure_logging('test')
    mock_constructor.assert_called_with('%(asctime)s - [%(name)s] - %(levelname)s %(module)s [%(pathname)s %(lineno)d]:\n%(message)s')
 
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above') 
def test_EnhancedStandardFormatter_prints_only_path_from_cwd():
    sys.pytl.log(f'hasattr(logging, "stream_handler") {hasattr(logging, "stream_handler")}')
    logging_file_path = pathlib.Path(configure_logging.STANDARD_LOGGING_LOCATION, 'test', configure_logging.EnhancedStdFormatter.LOG_FILENAME)
    logging_file_path.unlink(missing_ok=True)
    logging_file_path.parent.mkdir(parents=True, exist_ok=True)
    configure_logging.configure_logging('test')
    logger = logging.getLogger('test')
    logger.info('some bubbles')
    # NB caplog.text does not capture the **formatted** messages, only the messages as formatted by pytest's formatter
    # module_name = __name__ # 2022-08-07: suspicion: __name__ may be different in W10
    module_name = 'configure_logging'
    # the date-time part at the start should not contain a comma
    equals = configure_logging.EnhancedStdFormatter._LINE_OF_EQUALS
    # NB the ".*" at the start here was added 2022-08-18 because the logging path is first logged by configure_logging()
    pattern_str = fr'.*{equals}\n[\d\-\s:\.]+ - \[test\] - INFO {module_name} \[\s*(.+{module_name}.py) \d+\]:.*some bubbles\n{equals}.*'
    with open(logging_file_path, 'r') as f:
        log_file_text = f.read()
    result = re.match(pattern_str, log_file_text, re.DOTALL)
    assert result != None
    # now test to make sure that the path of the file where the log message was done, relative to the cwd root, is included
    file_path = pathlib.Path(configure_logging.__file__)
    last_part_of_path = file_path.relative_to(cwd_path.parent)
    
    assert result.groups()[0] == str(last_part_of_path)
 
def test_logging_dir_is_created_if_does_not_exist(tmpdir):
    app_name = cwd_path.parts[-1]
    logging_dir_path = pathlib.Path(str(tmpdir)).joinpath('logging_dir', app_name)
    logging_dir_path_str = str(logging_dir_path)
    with mock.patch('configure_logging.STANDARD_LOGGING_LOCATION', 
        new_callable=mock.PropertyMock(return_value=logging_dir_path_str)):
        configure_logging.configure_logging('test')
    assert logging_dir_path.is_dir()
 
def test_warn_message_out_if_logging_dir_fails_to_be_created(tmpdir):
    assert len(logging.getLogger().handlers) == 4
    app_name = cwd_path.parts[-1]
    logging_dir_path = pathlib.Path(str(tmpdir)).joinpath('logging_dir', app_name)
    logging_dir_path_str = str(logging_dir_path)
    with mock.patch('configure_logging.STANDARD_LOGGING_LOCATION', 
        new_callable=mock.PropertyMock(return_value=logging_dir_path_str)):
        with mock.patch('pathlib.Path.mkdir', side_effect=Exception()):
            with pytest.raises(Exception) as e:
                # sys.pytl.log(f'here')
                configure_logging.configure_logging('test')
                    
    # sys.pytl.log(f'here {e}')
    assert not logging_dir_path.is_dir()
    assert 'WARNING: failed to create logging directory' in str(e)
    # the custom console streamhandler *is* created, after removing the default stderr streamhandler
    assert len(logging.getLogger().handlers) == 4
 
@mock.patch(f'logging.basicConfig')
def test_basic_logger_is_not_configured_if_exception_raised(mock_basic):
    assert len(logging.getLogger().handlers) == 4
    with mock.patch('logging.Logger.setLevel', side_effect=Exception('random problem')):
        with pytest.raises(Exception):
            configure_logging.configure_logging('test')
    mock_basic.assert_not_called()
    # _, err = capsys.readouterr()
    # assert 'WARNING: unexpected problem trying to configure basic logger' in err
    # assert 'random problem' in err
     
def test_logger_info_message_with_logging_path_if_successful_configuration():
    with mock.patch('logging.Logger.info') as mock_info:
        configure_logging.configure_logging('test')
    logging_output_path = pathlib.Path(configure_logging.STANDARD_LOGGING_LOCATION, 'test', 
        configure_logging.EnhancedStdFormatter.LOG_FILENAME)
    # this got a "re.error: bad escape" before I did this
    logging_output_path = re.sub(r"\\", ".", str(logging_output_path))
    mock_info.assert_called_once_with(AnyStringWithRegex(f'.*{logging_output_path}.*'))
  
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists   
def test_do_start_run_checks_can_return_values_rather_than_exiting(_):
    app_name = cwd_path.parts[-1]
    VersionInfo = collections.namedtuple('VersionInfo', 'major minor micro')
    version_info = VersionInfo(3, 9, 1)
    with mock.patch('sys.version_info', version_info):
        return_value = library_main.do_start_run_checks(app_name, cwd_path.joinpath('src', 'core', '__main__.py'), return_value=True)
    assert return_value[0] == library_main.PYTHON_VERSION_IS_TOO_LOW     
 
@pytest.mark.parametrize('package_name', ['string_name', False, 0, None])
@pytest.mark.parametrize('return_vals', [True, False])
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
def test_do_start_run_checks_checks_package_name_param(_, package_name, return_vals, capsys):
    app_name = cwd_path.parts[-1]
    caller_file_path = cwd_path.joinpath('src', package_name, '__main__.py') if package_name == 'string_name' else pathlib.Path('blip') 
    if package_name == 'string_name':
        return_val = library_main.do_start_run_checks(app_name, caller_file_path, package_name=package_name, return_value=return_vals)
        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    elif return_vals:
        return_val = library_main.do_start_run_checks(app_name, caller_file_path, package_name=package_name, return_value=return_vals)
        assert return_val[0] == library_main.BAD_PARAMS
        assert re.match(fr'FATAL\..*Bad param.*', return_val[1]) != None, f'return_val[1] |{return_val[1]}|'
        _, err = capsys.readouterr()
        assert err == ''
    else:
        with pytest.raises(SystemExit) as excinfo:
            library_main.do_start_run_checks(app_name, caller_file_path, package_name=package_name, return_value=return_vals)
        assert excinfo.value.code == library_main.BAD_PARAMS
        _, err = capsys.readouterr()
        assert re.match(fr'FATAL\..*Bad param.*', err) != None, f'err |{err}|'
 
@pytest.mark.parametrize('return_vals', [True, False, 0, None, 'hi!'])
@mock.patch('pathlib.Path.is_file', return_value=True) # pretend that this __main__.py file exists
def test_do_start_run_checks_checks_return_vals_param(_, return_vals, capsys):
    app_name = cwd_path.parts[-1]
    caller_file_path = cwd_path.joinpath('src', 'core', '__main__.py')
    # NB 1 == True is True, and 0 == False is also True...
    if return_vals is True:
        return_val = library_main.do_start_run_checks(app_name, caller_file_path, return_value=return_vals)
        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    elif return_vals is False:
        return_val = library_main.do_start_run_checks(app_name, caller_file_path, return_value=return_vals)
        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    else:
        with pytest.raises(SystemExit) as excinfo:
            library_main.do_start_run_checks(app_name, caller_file_path, return_value=return_vals)
        assert excinfo.value.code == library_main.BAD_PARAMS           
        _, err = capsys.readouterr()
        assert re.match(fr'FATAL\..*Bad param.*', err) != None, f'err |{err}|'
         
@pytest.mark.parametrize('file_path', ['real', False, 0, None, 'bubbles'])
@pytest.mark.parametrize('return_vals', [True, False])
def test_do_start_run_checks_checks_caller_file_path_param(file_path, return_vals, capsys):
    app_name = cwd_path.parts[-1]
    if file_path == 'real':
        file_path = cwd_path.joinpath('src', 'core', '__main__.py')
        # pretend that this __main__.py file exists
        with mock.patch('pathlib.Path.is_file', return_value=True):
            return_val = library_main.do_start_run_checks(app_name, file_path, return_value=return_vals)
        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    elif return_vals:
        return_val = library_main.do_start_run_checks(app_name, file_path, return_value=return_vals)
        assert return_val[0] == library_main.BAD_PARAMS
        assert re.match(fr'FATAL\..*Bad param.*', return_val[1]) != None, f'return_val[1] |{return_val[1]}|'
        _, err = capsys.readouterr()
        assert err == ''
    else:
        with pytest.raises(SystemExit) as excinfo:
            library_main.do_start_run_checks(app_name, file_path, return_value=return_vals)
        assert excinfo.value.code == library_main.BAD_PARAMS
        _, err = capsys.readouterr()
        assert re.match(fr'FATAL\..*Bad param.*', err) != None, f'err |{err}|'
     
@pytest.mark.parametrize('app_name_param', ['real', False, 0, None])
@pytest.mark.parametrize('return_vals', [True, False])
def test_do_start_run_checks_checks_app_name_param(app_name_param, return_vals, capsys):
    app_name = cwd_path.parts[-1]
    caller_file_path = cwd_path.joinpath('src', 'core', '__main__.py')  
    if app_name_param == 'real':
        # use the real app_name, i.e. "lib"; also pretend that this __main__.py file exists
        with mock.patch('pathlib.Path.is_file', return_value=True):
            return_val = library_main.do_start_run_checks(app_name, caller_file_path, return_value=return_vals)
        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    elif return_vals:
        return_val = library_main.do_start_run_checks(app_name_param, caller_file_path, return_value=return_vals)
        assert return_val[0] == library_main.BAD_PARAMS
        assert re.match(fr'FATAL\..*Bad param.*', return_val[1]) != None, f'return_val[1] |{return_val[1]}|'
        _, err = capsys.readouterr()
        assert err == ''
    else:
        with pytest.raises(SystemExit) as excinfo:
            library_main.do_start_run_checks(app_name_param, caller_file_path, return_value=return_vals)
        assert excinfo.value.code == library_main.BAD_PARAMS
        _, err = capsys.readouterr()
        assert re.match(fr'FATAL\..*Bad param.*', err) != None, f'err |{err}|'
     
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')
@pytest.mark.parametrize('caller_file_path', ['file', 'dir', 'non-existent'])
@pytest.mark.parametrize('return_vals', [True, False])
def test_do_start_run_checks_checks_that_caller_file_path_is_file(caller_file_path, return_vals, capsys, tmpdir):
    app_name = cwd_path.parts[-1]
    tmpdir_path = pathlib.Path(str(tmpdir))
    bad_caller_file_path = tmpdir_path.joinpath('subdir')
    if caller_file_path == 'dir':
        bad_caller_file_path.mkdir()
    if caller_file_path == 'file':
        # this will be a real path, as long as this "lib" project has a package "library_core" and file "__init__.py" under it
        caller_file_path = cwd_path.joinpath('src', 'library_core', '__init__.py')
        # use the real app_name, i.e. "lib"
        return_val = library_main.do_start_run_checks(app_name, caller_file_path, package_name='library_core', return_value=return_vals)
        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    elif return_vals:
        return_val = library_main.do_start_run_checks(app_name, bad_caller_file_path, return_value=return_vals)
        assert return_val[0] == library_main.BAD_PARAMS
        assert re.match(fr'FATAL\..*Bad param.*', return_val[1]) != None, f'return_val[1] |{return_val[1]}|'
        _, err = capsys.readouterr()
        assert err == ''
    else:
        with pytest.raises(SystemExit) as excinfo:
            library_main.do_start_run_checks(app_name, bad_caller_file_path, return_value=return_vals)
        assert excinfo.value.code == library_main.BAD_PARAMS
        _, err = capsys.readouterr()
        assert re.match(fr'FATAL\..*Bad param.*', err) != None, f'err |{err}|'
 
@pytest.mark.parametrize('version_too_low', [True, False])
def test_version_check_message_if_lib_version_is_too_low(version_too_low, tmpdir):
    version_string = ('7.0.0' if version_too_low else '7.0.1')
    min_lib_version = packaging.version.Version('7.0.1') 

    # 2024-12-07 this has changed completely: library_main.lib_module_version_check must now read in file version.toml in the root dir

    # first read out to a temp dir a version.toml file with the "too-low version"
    tmpdir_path = pathlib.Path(str(tmpdir))
    too_low_version_dict = {'version': version_string}
    with open(tmpdir_path.joinpath('version.toml'), 'w') as f:
        toml.dump(too_low_version_dict, f)
    # mock out the CWD
    with mock.patch('pathlib.Path.cwd', return_value=tmpdir_path):

        # assert library_main.lib_module_version_check(min_lib_version) == (not version_too_low)
        assert library_main.lib_module_version_check(min_lib_version) == (not version_too_low)
 
@pytest.mark.skip(reason='2025-12-13 was working OK until I changed to using "import library_main" above')
@pytest.mark.parametrize('in_ve', [True, False])
@pytest.mark.parametrize('return_vals', [True, False])  
def test_main_stops_if_not_run_in_virtual_environment(capsys, in_ve, return_vals):
    assert sys.base_prefix != sys.prefix, 'it appears that these tests are not being run in a virtual environment'
    app_name = cwd_path.parts[-1]
    caller_file_path = cwd_path.joinpath('src', 'library_core', 'library_main.py')
    pattern_str = r'FATAL\..*This app is not being run in an appropriate activated Python virtual environment.*'
    if in_ve:
        return_val = library_main.do_start_run_checks(app_name, caller_file_path, package_name='library_core', return_value=return_vals)


        assert return_val == library_main.CHECKS_PASSED
        _, err = capsys.readouterr()
        assert err == ''
    else:
        with mock.patch('sys.base_prefix', new_callable=mock.PropertyMock(return_value=sys.prefix)):
            if return_vals:
                return_val = library_main.do_start_run_checks(app_name, caller_file_path, package_name='library_core', return_value=return_vals)
                error_code = return_val[0]
            else:
                with pytest.raises(SystemExit) as excinfo:
                    return_val = library_main.do_start_run_checks(app_name, caller_file_path, package_name='library_core', return_value=return_vals)
                error_code = excinfo.value.code
        _, err = capsys.readouterr()
        error_msg = return_val[1] if return_vals else err
        assert (err == '') if return_vals else True
        assert error_code == library_main.NOT_IN_VIRTUAL_ENVIRONMENT
        assert re.match(pattern_str, error_msg, flags=re.IGNORECASE) != None, f'error_msg |{error_msg}|'

@pytest.mark.parametrize('normal_configuration_succeeds', [True, False, 'exception'])
@mock.patch(f'logging.basicConfig')
def test_test_configure_logging_returns_True_or_False_or_raises_exception(mock_basic, capsys, normal_configuration_succeeds):
    assert len(logging.getLogger().handlers) == 4
    if normal_configuration_succeeds == True:
        return_value = configure_logging.configure_logging('test')
        assert return_value == True
    elif normal_configuration_succeeds == 'exception':
        with mock.patch('logging.Logger.setLevel', side_effect=Exception('random problem')):
            with pytest.raises(Exception):
                configure_logging.configure_logging('test')
    else:
        pass # TODO: something which results in False return
    
def test_console_StreamHandler_formatter_configures_legible_colours():
    configure_logging.configure_logging('test')
    logger = logging.getLogger('test')
    # NB a ConcurrentRotatingFileHandler subclasses from both FileHandler and StreamHandler
    console_stream_handlers = [handler for handler in logger.handlers \
        if isinstance(handler, colorlog.StreamHandler) and not isinstance(handler, logging.FileHandler)]
    assert len(console_stream_handlers) == 1
    console_stream_handler = console_stream_handlers[0]
    formatter = console_stream_handler.formatter
    log_colors={
        'DEBUG':    'white',
        'INFO':     'white',
        'WARNING':  'fg_bold_yellow',
        'ERROR':    'fg_bold_red',
        'CRITICAL': 'fg_bold_red,bg_white',
    }
    assert formatter.log_colors == log_colors
    
def test_configure_logger_sets_console_StreamHandler_only_for_the_first_logger_configured():
    assert not hasattr(logging, 'stream_handler')
    configure_logging.configure_logging('test1')
    logger1 = logging.getLogger('test1')
    assert hasattr(logging, 'stream_handler')
    # NB a ConcurrentRotatingFileHandler subclasses from both FileHandler and StreamHandler
    logger1_console_stream_handlers = [handler for handler in logger1.handlers \
        if isinstance(handler, colorlog.StreamHandler) and not isinstance(handler, logging.FileHandler)]
    assert len(logger1_console_stream_handlers) == 1
    # now create another logger, with a different name
    configure_logging.configure_logging('test2')
    logger2 = logging.getLogger('test2')
    assert hasattr(logging, 'stream_handler')
    logger2_console_stream_handlers = [handler for handler in logger2.handlers \
        if isinstance(handler, colorlog.StreamHandler) and not isinstance(handler, logging.FileHandler)]
    assert len(logger2_console_stream_handlers) == 0
        
         



import re 
    
class AnyStringWithRegex(str):
    def __init__(self, case_insensitive=True):
        self.case_insensitive = case_insensitive
    def __eq__(self, other):
        if self.case_insensitive:
            return len(re.findall(self.lower(), other.lower(), re.DOTALL)) != 0
        return len(re.findall(self, other, re.DOTALL)) != 0
    
def is_match(pattern, string, flags=re.IGNORECASE | re.DOTALL): # or "is_full_match", as desired
    return re.fullmatch(pattern, string, flags)!=None
